import sqlite3
import os
import time
import logging
from abc import abstractmethod

from .population import Population
from .other.sqlitedict.sqlitedict import SqliteDict


class SqliteHandler(logging.Handler):
    """
    Thread-safe logging handler for SQLite.
    """

    def __init__(self, data_store):
        logging.Handler.__init__(self)

        self.data_store = data_store
        self.data_store.create_structure_log()

    def emit(self, record: logging.LogRecord):
        """

        self.format(record)
        # format record
        record.dbtime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))
        if record.exc_info:  # for exceptions
            record.exc_text = logging._defaultFormatter.formatException(record.exc_info)
        else:
            record.exc_text = ""
        """

        # Insert the log record
        try:
            connection = sqlite3.connect(self.data_store.database_name)

            cursor = connection.cursor()
            cursor.execute("INSERT INTO log(timestamp, source, loglevel, loglevelname, message, args, module, funcname, lineno, exception, process, threadname) "
                           "VALUES (:timestamp, :source, :loglevel, :loglevelname, :message, :args, :module, :funcname, :lineno, :exception, :process, :threadname)",
                           { "timestamp": str(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))),
                             "source": str(record.name),
                             "loglevel": record.levelno,
                             "loglevelname": str(record.levelname),
                             "message": str(record.getMessage()),
                             "args": str(record.args),
                             "module": str(record.module),
                             "funcname": str(record.funcName),
                             "lineno": record.lineno,
                             "exception": str(record.exc_text),
                             "process": record.process,
                             "threadname": str(record.threadName) })
            connection.commit()
            cursor.close()

            connection.close()
        except sqlite3.Error as e:
            print("SqliteHandler: error occurred:", e.args[0])


class DataStore:
    """ Class  ensures saving data from optimization problems. """

    def __init__(self, problem):
        self.problem = problem

        # create the new loop and worker thread
        # self.worker_loop = asyncio.new_event_loop()
        # worker = Thread(target=self._start_worker, args=(self.worker_loop,))
        # worker.daemon = True
        # # Start the thread
        # worker.start()

        # populations
        self.populations = []

    def __del__(self):
        # print("DataStore:def __del__(self):")
        pass

    # def _start_worker(self, loop):
    #     asyncio.set_event_loop(loop)
    #     loop.run_forever()

    @abstractmethod
    def create_structure(self):
        pass

    @abstractmethod
    def write_individual(self, individual):
        # TODO: check this hack
        if len(self.populations) == 0:
            self.populations.append(Population())

        # add individual
        self.populations[-1].individuals.append(individual)

    @abstractmethod
    def write_population(self, population):
        self.populations.append(population)

    def get_id(self):
        return 0


class FileDataStore(DataStore):

    def __init__(self, problem, database_name=None, remove_existing=True, mode="write", backend="shelve"):
        super().__init__(problem)

        self.database_name = database_name
        self.mode = mode

        if remove_existing and mode == "write":
            if os.path.exists(self.database_name):
                os.remove(self.database_name)

        if backend == "shelve":
            import shelve

            if self.mode == "write":
                self.db = shelve.open(self.database_name, flag='c', writeback=True)
                # remove database and create structure
                if remove_existing and os.path.exists(self.database_name):
                    self.create_structure()
            else:
                self.db = shelve.open(self.database_name, flag='r')
                self.read_from_datastore()
        elif backend == "diskcache":
            import diskcache

            self.db = diskcache.Cache(directory=self.database_name)
            if self.mode == "write":
                # remove database and create structure
                if remove_existing and os.path.exists(self.database_name):
                    self.create_structure()
            else:
                self.read_from_datastore()
        elif backend == "sqlitedict":
            self.db = SqliteDict(self.database_name, autocommit=True)
            if self.mode == "write":
                # remove database and create structure
                if remove_existing and os.path.exists(self.database_name):
                    self.create_structure()
            else:
                self.read_from_datastore()
        else:
            assert 1

        # set datastore to problem
        self.problem.data_store = self

    def __del__(self):
        super().__del__()

    def create_structure(self):
        self.db["name"] = self.problem.name
        self.db["description"] = self.problem.description
        self.db["parameters"] = self.problem.parameters
        self.db["costs"] = self.problem.costs

        self.db["populations"] = []

    def write_individual(self, individual):
        # add individual
        super().write_individual(individual)

        if self.mode == "write":
            # TODO: check this hack
            populations = self.db["populations"]
            if len(populations) == 0:
                populations.append(Population())

            # add individual
            populations[-1].individuals.append(individual)
            self.db["populations"] = populations

    def write_population(self, population):
        super().write_population(population)

        if self.mode == "write":
            # write to database
            populations = self.db["populations"]
            populations.append(Population())
            self.db["populations"] = populations

    def read_from_datastore(self):
        self.problem.name = self.db["name"]
        self.problem.description = self.db["description"]
        self.problem.parameters = self.db["parameters"]
        self.problem.costs = self.db["costs"]

        self.populations = self.db["populations"]


class DummyDataStore(DataStore):

    def __init__(self, problem=None):
        super().__init__(problem)
        self.create_structure()

    def __del__(self):
        super().__del__()

    def create_structure(self):
        pass

    def write_individual(self, individual):
        super().write_individual(individual)

    def write_population(self, population):
        super().write_population(population)


#TODO: Create it from the Reference
"""
    Reference
    ---------
    Yew-Soon Ong, Member, IEEE, Prasanth B. Nair, and Kai Yew Lum
    "Max–Min Surrogate-Assisted Evolutionary Algorithm for Robust Design"
    IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 10, NO. 4, AUGUST 2006
"""


"""
Poloni (POL) function is a classic test function for multi-objective optimization methods as its Pareto optimal
solutions are not continous and non-convex.

References:
---------

[1] Lei, Gang; Zhu, Jianguo; Guo, Youguang, "Design Optimization Methods for Electrical Machines", 2016

[2] Lei G, Shao KR, Guo YG, Zhu JG (2012), "Multiobjective sequential optimization method for the design of
    industrial electromagnetic devices. IEEE Trans Magn 48(11):4538–4541"

section: 4.4.2
formula: 4.13 and 3.13
"""
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

import matplotlib.pyplot as plt
import artap.colormaps as cmaps
import numpy as np

from matplotlib.ticker import LinearLocator, FormatStrFormatter
from scipy import sin, cos


def func_POL(x1, x2):
    """

    :param x1: [-pi ; pi]
    :param x2: [-pi ; pi]
    :return:
    """
    A1 = 0.5 * sin(1.) - 2. * cos(1.) + sin(2.) - 1.5 * cos(2.)
    A2 = 1.5 * sin(1.) - cos(1.) + 2. * sin(2.) - 0.5 * cos(2.)

    B1 = 0.5 * sin(x1) - 2. * cos(x1) + sin(x2) - 1.5 * cos(x2)
    B2 = 1.5 * sin(x1) - cos(x1) + 2. * sin(x2) - 0.5 * cos(x2)

    f1 = 1. + (A1 - B1) ** 2. + (A2 - B2) ** 2.
    f2 = (x1 + 3.) ** 2. + (x2 + 1.) ** 2.

    return [f1, f2]


def plot_pol(pic):
    """
    pic = 0 -> plots the f1 function which has a maxima and two minima
    pic = 1 -> plots the f2 function which i a hyperbola in the given region
    """
    fig = plt.figure()
    ax = fig.gca(projection='3d')


# Make data.
    X = np.arange(-np.pi, np.pi, .1)
    Y = np.arange(-np.pi, np.pi, .1)
    X, Y = np.meshgrid(X, Y)

    Z = func_POL(X, Y)
    # Plot the surface.
    surf = ax.plot_surface(X, Y, Z[pic], cmap=cmaps.viridis,
                           linewidth=0, antialiased=True)

    # Customize the z axis.
    # ax.set_zlim(-1.01, 1.01)
    ax.zaxis.set_major_locator(LinearLocator(10))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()

if __name__ == "__main__":
    plot_pol(0)
    plot_pol(1)


"""
IMPLEMENTATION of BENCHMARK FUNCTIONS from CEC2020:
---------------------------------------------------

Referecces:

[1] https://github.com/P-N-Suganthan/2020-Multimodal-Multi-Objective-Benchmark
[2] J. J. Liang1, P. N. Suganthan2, B. Y. Qu3, D. W. Gong4 and C. T. Yue1,
    "Problem Definitions and Evaluation Criteria for the CEC 2020 Special Session on Multimodal Multiobjective
    Optimization"

"""
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

import matplotlib.pyplot as plt
import artap.colormaps as cmaps
from matplotlib.ticker import LinearLocator, FormatStrFormatter

from numpy import vectorize
from numpy import pi, sin, e, cos
import numpy as np


def MMF1(x1, x2):
    """
    Properties
    =================================================================
    Scalable number of variables :        x
    Scalable number of objectives:        x
    Pareto optima known:                  yes
    Pareto front geometry:                convex
    Pareto set geometry:                  nonlinear
    Scalable number of Pareto set:        x
    Nr of global optimums:                2
    NR of local optimums:                 1
    =================================================================

    Solution
    --------
    global pareto set:

    x1 = x1
    x2 = sin(6*pi*abs(x1-2)+pi)

    global pareto front:    f2 =  1 - sqrt(f1) , where 0<=f1<=1

    :param x1: [1,3]
    :param x2: [-1,1]
    :return: [f1, f2]
    """

    # a = e  # this can be any other number, the search space of x2 is [-a^3, a^3]
    # if x1 < 2.:
    #    a = 1.
    a = np.where(x1 < 2., 1., e)
    f1 = abs(x1 - 2)
    f2 = 1 - abs(x1 - 2) ** 0.5 + 2. * (x2 - (a ** x1) * sin(6. * pi * abs(x1 - 2.) + pi)) ** 2.
    return [f1, f2]


def MMF2(x1, x2):
    """
    Properties
    =================================================================
    Scalable number of variables :        x
    Scalable number of objectives:        x
    Pareto optima known:                  yes
    Pareto front geometry:                convex
    Pareto set geometry:                  nonlinear
    Scalable number of Pareto set:        x
    Nr of global optimums:                2
    NR of local optimums:                 1
    =================================================================

    Solution
    --------
    global pareto set:

    x1 = x2**2  if x1<1
         (x2-1) if 1<x1<2
    x2 = x2

    global pareto front:    f2 =  1 - sqrt(f1) , where 0<=f1<=1

    :param x1: [0,1]
    :param x2: [0,2]
    :return: [f1, f2]
    """

    f1 = abs(x1)

    a = x1 ** 0.5

    f2a = 1 - a + 2. * (4. * (x2 - a)) ** 2. - 2. * cos(20.0 * pi / (2.0 ** 0.5) * (x2 - a)) + 2.
    f2b = 1 - a + 2. * (4. * (x2 - 1.0 - a)) ** 2. - 2. * cos(20.0 * pi / (2.0 ** 0.5) * (x2 - 1.0 - a)) + 2.

    f2 = np.where(x2 < 1., f2a, f2b)

    return [f1, f2]


def plot_MMF1(selector):
    """
    :param selector: 0 or 1 to plot f1 or f2
    :return:
    """
    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # Make data.
    X = np.arange(1., 3., .01)
    Y = np.arange(-e, e, .01)
    X, Y = np.meshgrid(X, Y)

    Z = MMF1(X, Y)
    # Plot the surface.
    surf = ax.plot_surface(X, Y, Z[selector], cmap=cmaps.viridis,
                           linewidth=0, antialiased=True)

    # Customize the z axis.
    ax.zaxis.set_major_locator(LinearLocator(10))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()


def plot_MMF2(selector):
    """
    :param selector: 0 or 1 to plot f1 or f2
    :return:
    """
    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # Make data.
    X = np.arange(0., 1., .01)
    Y = np.arange(0., 2., .01)
    X, Y = np.meshgrid(X, Y)

    Z = MMF2(X, Y)

    # Plot the surface.
    surf = ax.plot_surface(X, Y, Z[selector], cmap=cmaps.viridis, linewidth=0, antialiased=True)

    # Customize the z axis.
    ax.zaxis.set_major_locator(LinearLocator(10))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()


if __name__ == "__main__":
    plot_MMF1(0)
    plot_MMF1(1)

    #plot_MMF2(0)



# from .algorithm import *
# from .executor import *
# from .algorithm_scipy import *
# from .algorithm_genetic import *

# import platform
#
# __agros__ = True
#
# try:
#     from agrossuite import agros as a2d
#
# except ImportError:
#     __agros__ = False
#
# try:
#     from .algorithm_bayesopt import *_
# except ImportError:
#     __bayes_opt__ = False
#
# try:
#     from .algorithm_nlopt import *_
# except ImportError:
#     __nl_opt__ = False
#
# # TODO: test if local Comsol is present
#
# # TODO: Repare
# if platform.system() == 'Linux':
#     from .algorithm_bayesopt import *
#
#     __all__ = ['Problem', 'NSGA_II', 'BayesOpt']
#
# else:
#     __all__ = ['Problem', 'NSGA_II']
# __all__ = ['Problem', 'NSGA_II']


import textwrap
import re
import rpyc
from rpyc.core.async_ import AsyncResultTimeout
from rpyc.utils.classic import upload_file, download_file
import asyncio
import tempfile
import os
import datetime
import time
import ntpath
from string import Template
from xml.dom import minidom
from artap.config import config

from abc import ABCMeta, abstractmethod
from .utils import ConfigDictionary
from shutil import copyfile


# parse ip address
def parse_address(address):
    regex_match = re.compile(r'<(\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}):.*').match(address)
    if regex_match is not None:
        return regex_match.group(1)
    return ""


class Executor(metaclass=ABCMeta):

    """
    Function is a class representing objective or cost function for
    optimization problems.
    """

    def __init__(self, problem):
        self.problem = problem

        self.options = ConfigDictionary()
        # parse method
        self.parse_results = None

    @abstractmethod
    def eval(self, x):
        # set default parse method from problem (can be overridden)
        if self.parse_results is None and "parse_results" in dir(self.problem):
            self.parse_results = self.problem.parse_results

    @staticmethod
    def _join_parameters_names(parameters, sep=","):
        param_names_string = ""

        for parameter in parameters:
            param_names_string += str(parameter['name']) + sep

        # remove last sep
        if (len(parameters)) > len(sep):
            param_names_string = param_names_string[:-len(sep)]

        return param_names_string

    @staticmethod
    def _join_parameters_values(x, sep=","):
        param_values_string = ""
        for number in x:
            param_values_string += str(number) + sep

        # remove last sep
        if (len(x)) > len(sep):
            param_values_string = param_values_string[:-len(sep)]

        return param_values_string

    @staticmethod
    def _remove_dir(local_dir):
        files = os.listdir(path=local_dir)
        for f in files:
            filepath = os.path.join(local_dir, f)
            os.remove(filepath)

        os.rmdir(local_dir)


class LocalComsolExecutor(Executor):

    comsol_command = Template("comsol batch -inputfile $input_file -nosave -pname $param_names -plist $param_values")

    def __init__(self, problem, problem_file, input_files=None, output_files=None):
        super().__init__(problem)
        self.model_file = ntpath.basename(problem_file)
        copyfile(problem_file, self.problem.working_dir + self.model_file)

        self.output_files = output_files
        self.input_files = input_files

    def eval(self, individual):
        super().eval(individual)

        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)

        run_string = self.comsol_command.substitute(input_file=self.model_file,
                                                    param_names=param_names_string,
                                                    param_values=param_values_string)

        # run command
        current_path = os.getcwd()
        os.chdir(self.problem.working_dir)
        os.system(run_string)
        os.chdir(current_path)
        output_files = []
        for file in self.output_files:
            output_files.append(self.problem.working_dir + file)
        result = self.parse_results(output_files, individual)
        return result


class RemoteExecutor(Executor):
    """
        Allows distributing of calculation of objective functions.
        """

    # ToDo: Make inheritated classes: Matlab, Comsol ...

    def __init__(self, problem, command, files_to_server, files_from_server=None):
        super().__init__(problem)

        if problem.working_dir is None:
            raise Exception('RemoteExecutor: Problem working directory must be set.')

        self.options.declare(name='hostname', default=config["condor_host"],
                             desc='Hostname')
        self.options.declare(name='username', default=config["condor_login"],
                             desc='Username')
        self.options.declare(name='port', default=15900, lower=0,
                             desc='Port')

        # command
        self.command = command

        # files
        self.input_files = files_to_server
        self.output_files = files_from_server

    @abstractmethod
    def eval(self, individual):
        super().eval(individual)

    def _create_client(self):
        # key = os.path.join(os.path.dirname(__file__), "cert/artap.key")
        # cert = os.path.join(os.path.dirname(__file__), "cert/artap.crt")

        channel = rpyc.Channel(rpyc.SocketStream.connect(self.options["hostname"], self.options["port"]))
        channel.send(self.options["username"].encode('utf-8'))
        response = channel.recv()
        AUTH_ERROR = b'error'
        if response == AUTH_ERROR:
            raise rpyc.utils.authenticators.AuthenticationError('Invalid username for daemon')

        return rpyc.utils.factory.connect_channel(channel, service=rpyc.core.ClassicService)
        # return rpyc.classic.connect(self.options["hostname"], self.options["port"])
        # return rpyc.classic.ssl_connect(self.options["hostname"], self.options["port"], keyfile=key, certfile=cert)

    def _init_remote(self, client):
        try:
            # create client
            remote_dir = client.root.create_job_dir()
            return remote_dir
        except RuntimeError as e:
            #self.problem.logger.error("Cannot create remote directory '{}' on ''".format(directory,
            #                                                                             self.options["hostname"]))
            print("ERROR --------------------------------")
            pass

    @staticmethod
    def _create_file_on_remote(destination_file, content, remote_dir, client):
        fp = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
        fp.write(content)
        fp.close()

        RemoteExecutor._transfer_file_to_remote(fp.name, destination_file, remote_dir, client)

        os.unlink(fp.name)

    @staticmethod
    def _transfer_file_to_remote(source_file, destination_file, remote_dir, client):
        source = source_file
        dest = remote_dir + os.sep + destination_file

        upload_file(client, localpath=source, remotepath=client.root.artap_dir + os.sep + dest)

    @staticmethod
    def _transfer_file_from_remote(source_file, destination_file, remote_dir, client):
        dest = destination_file
        source = remote_dir + os.sep + source_file

        download_file(client, remotepath=client.root.artap_dir + os.sep + source, localpath=dest)

    @staticmethod
    def _create_dir_on_remote(directory, client):
        d = datetime.datetime.now()
        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")

        # projects directory
        client.exec_command("mkdir " + directory)

        # working directory
        wd = directory + os.sep + "artap-" + ts
        client.exec_command("mkdir " + wd)

        return wd

    def _transfer_files_to_remote(self, remote_dir, client):
        # transfer input files
        if self.input_files:
            for file in self.input_files:
                self._transfer_file_to_remote(self.problem.working_dir + os.sep + file, "." + os.sep + file,
                                              remote_dir=remote_dir, client=client)

    def _transfer_files_from_remote(self, client):
        pass


class CondorJobExecutor(RemoteExecutor):
    """
    Allows distributing of calculation of objective functions.
    """
    def __init__(self, problem, files_to_condor, files_from_condor=None):
        super().__init__(problem, "", files_to_condor, files_from_condor)

        # set default host
        self.options["hostname"] = config["condor_host"]

    @abstractmethod
    def _create_job_file(self, remote_dir, individual, client):
        pass

    def eval(self, individual):
        if config["condor_host"] is None:
            raise Exception("Condor host is not defined.")

        super().eval(individual)

        success = False
        while not success:
            try:
                # create client
                client = self._create_client()

                # init remote
                remote_dir = self._init_remote(client=client)
                # self.problem.logger.info("RemoteDir {}".format(remote_dir))

                # transfer supplementary files, input and model file
                self._transfer_files_to_remote(remote_dir, client)

                # submit job
                # print("_create_job_file -      {}".format(remote_dir))
                try:
                    self._create_job_file(remote_dir, individual, client)
                except AsyncResultTimeout as e:
                    print("ERROR - {} - try again ({})".format(e, remote_dir))
                    # remove job dir
                    client.root.remove_job_dir(remote_dir)
                    client.close()
                    # resubmit
                    return self.eval(individual)

                start = time.time()

                events = []
                cnt = 0
                return_value = -1
                delay = 0.5

                run = True
                while run:
                    eventlog = client.root.eventlog(remote_dir)
                    for e in eventlog:
                        if {e.timestamp, e.type} not in events:
                            events.append({e.timestamp, e.type})

                    # print("len(events) = {}".format(len(events)))
                    if len(events) > 0:
                        for i in range(cnt, len(events)):
                            event = eventlog[i]
                            tp = int(event.type)

                            args = ""

                            if tp == 0:
                                # SUBMIT
                                if "ReturnValue" in event:
                                    return_value = event["ReturnValue"]
                            elif tp == 1:
                                # EXECUTE
                                if "ExecuteHost" in event:
                                    args += "ExecuteHost: {}, ".format(parse_address(event["ExecuteHost"]))
                            elif tp == 6:
                                # IMAGE_SIZE
                                pass
                            elif tp == 5:
                                # JOB_TERMINATED
                                if "ReturnValue" in event:
                                    return_value = event["ReturnValue"]
                                run = False
                            elif tp == 12:
                                # JOB_HELD
                                if "ReturnValue" in event:
                                    return_value = event["ReturnValue"]

                                self.problem.logger.error("Job {} is '{}' at {}".format(event.type, event.type, event.type))
                                run = False
                                # read log
                                #content_log = self._read_file_from_remote("{}.log".format(self.output_files[0]),
                                #                                          remote_dir=remote_dir, client=client)
                                #self.problem.logger.error(content_log)
                                # remove job
                                # self._run_command_on_remote("condor_rm {}".format(process_id),
                                #                            remote_dir=remote_dir, client=client)
                                raise RuntimeError

                            if len(args) > 0:
                                args = args[:-2]

                            self.problem.logger.info("Job {}.{} ({}) is '{}' at {}".format(event.cluster, event.proc, remote_dir, event.type, args))
                            # print("{}: {} ({})".format(eventlog[i].timestamp, eventlog[i].type, args))

                    if run:
                        cnt = len(events)
                        time.sleep(delay)

                end = time.time()
                if (end - start) > self.problem.options["time_out"]:
                    raise TimeoutError

                # successfull
                if return_value == 0:
                    if len(self.output_files) > 0:
                        output_files = []
                        d = datetime.datetime.now()
                        ts = d.strftime("%Y-%m-%d-%H-%M-%S-%f")
                        path = self.problem.working_dir + 'artap' + ts
                        os.mkdir(path)

                        for file in self.output_files:
                            self._transfer_file_from_remote(source_file=file, destination_file=path + os.sep + file,
                                                            remote_dir=remote_dir, client=client)
                            output_files.append(path + os.sep + file)
                        success = True
                        result = self.parse_results(output_files, individual)

                    # remove job dir
                    if result is not None:
                        client.root.remove_job_dir(remote_dir)

                    if self.problem.options['save_data_files'] is False:
                        self._remove_dir(path)
                else:
                    assert 0

                # remove job dir
                # client.root.remove_job_dir(remote_dir)

                client.close()

                return result

            except ConnectionError as e:
                print(e)
                time.sleep(1.0)
                continue


class CondorPythonJobExecutor(CondorJobExecutor):
    executable = textwrap.dedent("""\
        #!/bin/sh
        # args

        python3 $@""")

    def __init__(self, problem, script, parameter_file, output_files=None):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], output_files)
        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = self.output_files
        condor_input_files = [self.script]
        # create input file with parameters

        if self.parameter_file:
            # parameters
            # create remote file
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            condor_input_files.append(self.parameter_file)
            arguments = self.script + " " + self.parameter_file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir,
                                        client=client)
        else:
            param_values_string = Executor._join_parameters_values(individual.vector, ",")
            arguments = self.script + " " + param_values_string

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        client.root.submit_job(remote_dir=remote_dir,
                               executable=client.root.artap_dir + os.sep + remote_dir + os.sep + "run.sh",
                               arguments=arguments,
                               input_files=condor_input_files,
                               output_files=condor_output_files)


class CondorMatlabJobExecutor(CondorJobExecutor):
    executable = textwrap.dedent("""\
        #!/bin/sh
        # args
        s=$@

        /opt/matlab-R2018b/bin/matlab -nodisplay -nosplash -nodesktop -r ${s%.m}""")

    def __init__(self, problem, script, parameter_file, files_from_condor=None):
        self.script = ntpath.basename(script)
        super().__init__(problem, [self.script], files_from_condor)
        self.parameter_file = parameter_file
        copyfile(script, self.problem.working_dir + self.script)

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = self.output_files
        condor_input_files = [self.parameter_file, self.input_files[0]]
        # create input file with parameters

        if self.parameter_file:
            # parameters
            param_values_string = Executor._join_parameters_values(individual.vector, "\n")
            # create remote file
            self._create_file_on_remote(self.parameter_file, param_values_string, remote_dir=remote_dir, client=client)

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        client.root.submit_job(remote_dir=remote_dir,
                               executable=client.root.artap_dir + os.sep + remote_dir + os.sep + "run.sh",
                               arguments=self.script,
                               input_files=condor_input_files,
                               output_files=condor_output_files)


class CondorComsolJobExecutor(CondorJobExecutor):
    arguments = Template("-inputfile $input_file -nosave -pname $param_names -plist $param_values")
    executable = textwrap.dedent("""\
            #!/bin/sh
            /opt/comsol-5.4/bin/comsol batch $@
            """)

    def __init__(self, problem, model_file, files_from_condor=None):
        self.model_file = ntpath.basename(model_file)
        super().__init__(problem, [self.model_file], files_from_condor)
        copyfile(model_file, self.problem.working_dir + self.model_file)

    def _create_job_file(self, remote_dir, individual, client):
        condor_output_files = self.output_files
        param_names_string = Executor._join_parameters_names(self.problem.parameters)
        param_values_string = Executor._join_parameters_values(individual.vector)
        arguments = self.arguments.substitute(input_file=os.path.basename(self.model_file),
                                              param_names=param_names_string,
                                              param_values=param_values_string)

        # create executable
        self._create_file_on_remote("run.sh", self.executable, remote_dir=remote_dir, client=client)

        client.root.submit_job(remote_dir=remote_dir,
                               executable=client.root.artap_dir + os.sep + remote_dir + os.sep + "run.sh",
                               arguments=arguments,
                               input_files=[self.model_file],
                               output_files=condor_output_files,
                               request_cpus=2,
                               request_memory=30)



import unittest
from artap.benchmarks.synthetic_functions import eval_5d_synthetic, eval_10d_synthetic


class TestSynthetic(unittest.TestCase):
    """ Tests for the synthetic benchmark functions """

    def test_synthetic5d(self):
        self.assertAlmostEqual(eval_5d_synthetic([10., 1.0, 6.0, 7.0, 8.0]), .7)
        self.assertAlmostEqual(eval_5d_synthetic([1.0, 3.0, 8.0, 9.5, 2.0]), .75)
        self.assertAlmostEqual(eval_5d_synthetic([3.0, 1.0, 3.0, 2.0, 5.0]), 1.0)
        self.assertAlmostEqual(eval_5d_synthetic([3.0, 4.0, 1.3, 5.0, 5.0]), 1.2, 5)

        self.assertAlmostEqual(eval_5d_synthetic([5.0, 2.0, 9.6, 7.3, 8.6]), 1.0)
        self.assertAlmostEqual(eval_5d_synthetic([7.5, 8.0, 9.0, 3.2, 4.6]), .6, 4)
        self.assertAlmostEqual(eval_5d_synthetic([5.7, 9.3, 2.2, 8.4, 7.1]), .5)
        self.assertAlmostEqual(eval_5d_synthetic([5.5, 7.2, 5.8, 2.3, 4.5]), .2, 4)

        self.assertAlmostEqual(eval_5d_synthetic([4.7, 3.2, 5.5, 7.1, 3.3]), 0.4)
        self.assertAlmostEqual(eval_5d_synthetic([9.7, 8.4, 0.6, 3.2, 8.5]), 0.1)

    def test_synthetic10d(self):
        self.assertAlmostEqual(eval_10d_synthetic([10., 1.0, 6.0, 7.0, 8.0, 1.0, 1.0, 6.0, 7.0, 8.0]), 0.7)
        self.assertAlmostEqual(eval_10d_synthetic([1.0, 3.0, 8.0, 9.5, 2.0, 1.0, 3.0, 8.0, 9.5, 2.0]), 0.75)
        self.assertAlmostEqual(eval_10d_synthetic([3.0, 1.0, 3.0, 2.0, 5.0, 3.0, 1.0, 3.0, 2.0, 5.0]), 1.0)
        self.assertAlmostEqual(eval_10d_synthetic([3.0, 4.0, 1.3, 5.0, 5.0, 3.0, 4.0, 1.3, 5.0, 5.0]), 1.2)


if __name__ == '__main__':
    unittest.main()



"""
This file contains the implementation of the following benchmark functions::

    - 1D synthetic test function for rubust design optimization (RDO)
    - 2D synthetic test function for RDO
    - 5D synthetic test function for RDO

    - Ackley test function

Definition and a plot of the 2D synthetic function for testing the robustness.
"""
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

import matplotlib.pyplot as plt
import artap.colormaps as cmaps
from matplotlib.ticker import LinearLocator, FormatStrFormatter
from numpy import exp
import numpy as np


def eval_2d_synthetic(x1, x2):
    """
    Gaussian 2D synthetic test function for testing the robustness.

    Peaks:
    -----
    (1,1) - width 0.3, multiplier 0.7
    (1,3) - width 0.4, multiplier 0.75
    (3,1)*- width 1.0, multiplier 1.0  <- robust solution
    (3,4) - width 0.4, multiplier 1.2  <- global optimum
    (5,2) - width 0.6, multiplier 1.0

    The function is defined on the [0,5], [0,5] region.

    Reference:
    ---------
    Yew-Soon Ong, Member, IEEE, Prasanth B. Nair, and Kai Yew Lum
    "Max–Min Surrogate-Assisted Evolutionary Algorithm for Robust Design"
    IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 10, NO. 4, AUGUST 2006
    """

    res = 0.7 * exp(-((x1 - 1.) ** 2. + (x2 - 1.) ** 2.) / 0.18) + \
          0.75 * exp(-((x1 - 1.) ** 2. + (x2 - 3.) ** 2.) / 0.32) + \
          exp(-((x1 - 3) ** 2 + (x2 - 1) ** 2) / 2.) + \
          1.2 * exp(-((x1 - 3) ** 2 + (x2 - 4) ** 2) / 0.32) + \
          exp(-((x1 - 5) ** 2. + (x2 - 2) ** 2.) / 0.72)

    return res


def plot_2d_synthetic():
    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # Make data.
    X = np.arange(0, 5, .01)
    Y = np.arange(0, 5, .01)
    X, Y = np.meshgrid(X, Y)

    # print(eval_2d_synthetic(X,Y))
    print(eval_2d_synthetic(X, Y))

    Z = eval_2d_synthetic(X, Y)  # np.sqrt(X**2 + Y**2)

    # Plot the surface.
    surf = ax.plot_surface(X, Y, Z, cmap=cmaps.viridis,
                           linewidth=0, antialiased=False)

    # Customize the z axis.
    # ax.set_zlim(-1.01, 1.01)
    ax.zaxis.set_major_locator(LinearLocator(10))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()


def eval_1d_synthetic(y: list):
    """
    This function was introduced to test the MEM - multiple-evaluation model schema, to compare it with the single
    evaluation model.

    Reference:
    ---------
    Yew-Soon Ong, Member, IEEE, Prasanth B. Nair, and Kai Yew Lum
    "Max–Min Surrogate-Assisted Evolutionary Algorithm for Robust Design"
    IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 10, NO. 4, AUGUST 2006
    """

    x = y[0]

    res = exp(-(x - 1) ** 2. / 0.5) + 2. * exp(-(x - 1.25) ** 2. / 0.045) + 0.5 * exp(-(x - 1.5) ** 2. / 0.0128) + \
          2. * exp(-(x - 1.6) ** 2. / 0.005) + 2.5 * exp(-(x - 1.8) ** 2. / 0.02) + \
          2.5 * exp(-(x - 2.2) ** 2. / 0.02) + 2. * exp(-(x - 2.4) ** 2. / 0.005) + \
          2. * exp(-(x - 2.75) ** 2. / 0.045) + exp(-(x - 3) ** 2. / 0.5) + 2. * exp(-(x - 6.) ** 2. / 0.32) + \
          2.2 * exp(-(x - 7.) ** 2. / 0.18) + 2.4 * exp(-(x - 8.) ** 2. / 0.5) + \
          2.3 * exp(-(x - 9.5) ** 2. / 0.5) + 3.2 * exp(-(x - 11.) ** 2. / 0.18) + 1.2 * exp(-(x - 12.) ** 2. / 0.18)

    return res


def runningMeanFast(x, N):
    return np.convolve(x, np.ones((N,)) / N)[(N - 1):]


def plot_1d_synthetic():
    x = np.arange(0.0, 12.0, 0.1)

    plt.figure()
    plt.plot(x, eval_1d_synthetic([x]), 'black',
             x, runningMeanFast(eval_1d_synthetic([x]), 10), '--')
    plt.xlabel("x")
    plt.ylabel("f(x)")
    plt.grid(True)
    plt.legend(['f(x)', 'avg(f(x))'])
    plt.show()


def atom_nd(width, multiplier, x: list, z: list):
    """
    Atomic function to generate an n-dimensional gaussian test function.

    multiplier: the amplitude/peak value of the function
    width: is the divider
    z: represents the zeros of the function
    x: represents the variables
    """
    res = 0
    for i in range(0, len(x)):
        res += (x[i] - z[i]) ** 2.

    res /= -width

    return exp(res) * multiplier


def eval_5d_synthetic(x: list):
    """
    Evaluates a 5 dimensional Gaussian - test function

    Reference:
    ---------
    Yew-Soon Ong, Member, IEEE, Prasanth B. Nair, and Kai Yew Lum
    "Max–Min Surrogate-Assisted Evolutionary Algorithm for Robust Design"
    IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 10, NO. 4, AUGUST 2006
    """
    result = atom_nd(0.3, 0.7, x, [10., 1.0, 6.0, 7.0, 8.0])
    result += atom_nd(0.4, 0.75, x, [1.0, 3.0, 8.0, 9.5, 2.0])
    result += atom_nd(1.0, 1.0, x, [3.0, 1.0, 3.0, 2.0, 5.0])  # robust solution
    result += atom_nd(0.4, 1.2, x, [3.0, 4.0, 1.3, 5.0, 5.0])  # highest peak
    result += atom_nd(0.6, 1.0, x, [5.0, 2.0, 9.6, 7.3, 8.6])
    result += atom_nd(0.5, 0.6, x, [7.5, 8.0, 9.0, 3.2, 4.6])
    result += atom_nd(0.1, 0.5, x, [5.7, 9.3, 2.2, 8.4, 7.1])
    result += atom_nd(1.0, 0.2, x, [5.5, 7.2, 5.8, 2.3, 4.5])
    result += atom_nd(0.2, 0.4, x, [4.7, 3.2, 5.5, 7.1, 3.3])
    result += atom_nd(0.3, 0.1, x, [9.7, 8.4, 0.6, 3.2, 8.5])

    return result


def eval_10d_synthetic(x: list):
    """
    Evaluates a 10 dimensional Gaussian - test function

    Reference:
    ---------
    Yew-Soon Ong, Member, IEEE, Prasanth B. Nair, and Kai Yew Lum
    "Max–Min Surrogate-Assisted Evolutionary Algorithm for Robust Design"
    IEEE TRANSACTIONS ON EVOLUTIONARY COMPUTATION, VOL. 10, NO. 4, AUGUST 2006
    """
    result = atom_nd(0.3, 0.7, x, [10., 1.0, 6.0, 7.0, 8.0, 1.0, 1.0, 6.0, 7.0, 8.0])
    result += atom_nd(0.4, 0.75, x, [1.0, 3.0, 8.0, 9.5, 2.0, 1.0, 3.0, 8.0, 9.5, 2.0])
    result += atom_nd(1.0, 1.0, x, [3.0, 1.0, 3.0, 2.0, 5.0, 3.0, 1.0, 3.0, 2.0, 5.0])  # robust solution
    result += atom_nd(0.4, 1.2, x, [3.0, 4.0, 1.3, 5.0, 5.0, 3.0, 4.0, 1.3, 5.0, 5.0])  # highest peak
    result += atom_nd(0.6, 1.0, x, [5.0, 2.0, 9.6, 7.3, 8.6, 5.0, 2.0, 9.6, 7.3, 8.6])
    result += atom_nd(0.5, 0.6, x, [7.5, 8.0, 9.0, 3.2, 4.6, 7.5, 8.0, 9.0, 3.2, 4.6])
    result += atom_nd(0.1, 0.5, x, [5.7, 9.3, 2.2, 8.4, 7.1, 5.7, 9.3, 2.2, 8.4, 7.1])
    result += atom_nd(1.0, 0.2, x, [5.5, 7.2, 5.8, 2.3, 4.5, 5.5, 7.2, 5.8, 2.3, 4.5])
    result += atom_nd(0.2, 0.4, x, [4.7, 3.2, 5.5, 7.1, 3.3, 4.7, 3.2, 5.5, 7.1, 3.3])
    result += atom_nd(0.3, 0.1, x, [9.7, 8.4, 0.6, 3.2, 8.5, 9.7, 8.4, 0.6, 3.2, 8.5])

    return result


def eval_ackley(x: list):
    """
    Implementation is based on: https://en.wikipedia.org/wiki/Ackley_function
    The minimum is placed in [0,0]
    """
    return -20. * exp(-0.2 * (0.5 * (x[0] ** 2 + x[1] ** 2)) ** 0.5) - \
           exp(0.5 * (np.cos(2. * np.pi * x[0]) + np.cos(2. * np.pi * x[1]))) + np.e + 20.


def plot_2d_ackley():
    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # Make data.
    X = np.arange(-32, 32, .1)
    Y = np.arange(-32, 32, .1)
    X, Y = np.meshgrid(X, Y)

    Z = eval_ackley([X, Y])
    # Plot the surface.
    surf = ax.plot_surface(X, Y, Z, cmap=cmaps.viridis,
                           linewidth=0, antialiased=True)

    # Customize the z axis.
    # ax.set_zlim(-1.01, 1.01)
    ax.zaxis.set_major_locator(LinearLocator(10))
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()


if __name__ == "__main__":
    plot_2d_synthetic()
    plot_1d_synthetic()
    plot_2d_ackley()
